#############################################################################
# Software Testing Automation Framework (STAF)                              #
# (C) Copyright IBM Corp. 2001                                              #
#                                                                           #
# This software is licensed under the Eclipse Public License (EPL) V1.0.    #
#############################################################################

###############################################################################
# THIS MODULE IS REQUIRED BY ALL PERL APPS THAT WANT TO INTERACT WITH STAF    #
# AND SHOULD NOT BE MODIFIED.                                                 #
###############################################################################

package PLSTAFPP;

use 5.008; # utf support
use strict;
use warnings;

our $VERSION = '3.0';

use constant BUFSIZE => 8192;  # size of each read of a STAF reply

use IO::Socket::UNIX qw();
use Encode           qw(encode_utf8 decode_utf8);

# (require "bytes.pm" routines, but for use with bytes:: prefix, so
# don't import "bytes" behavior by default)
no bytes;

sub STAF::Register
{
    if (@_ != 1)
    {
        die "Usage: STAF::Register(processName)\n";
    }

    my $process_name = shift;
    ($STAF::RC, $STAF::Result, $STAF::Handle) = _transact
        (process_name => $process_name,
         pid          => $$,
         encoder      => \&_encode_register,
         decoder      => \&_decode_register);
    return $STAF::RC;
}

sub STAF::UnRegister
{
    if (@_ != 0)
    {
        die "Usage: STAF::UnRegister()\n";
    }
    ($STAF::RC, $STAF::Result) = _transact
        (handle  => $STAF::Handle,
         pid     => $$,
         encoder => \&_encode_unregister,
         decoder => \&_decode_unregister);
    return $STAF::RC;
}

sub STAF::Submit
{
    if (@_ < 3 || @_ > 4)
    {
        die "Usage: STAF::Submit(location, service, request, syncoption=kSTAFReqSync)\n";
    }
    my ($location, $service, $request, $syncoption) = @_;
    $syncoption ||= $STAF::STAFHandle::kReqSync;

    ($STAF::RC, $STAF::Result) = _transact
        (location   => $location,
         service    => $service,
         request    => $request,
         syncoption => $syncoption,
         pid        => $$,
         handle     => $STAF::Handle,
         encoder    => \&_encode_submit,
         decoder    => \&_decode_submit);
    return $STAF::RC;
}

sub _encode_register
{
    my (%opts) = @_;
    pack('VVVV',
         2, # API number
         0, # API version
         $opts{pid},
         bytes::length($opts{process_name}))
        . encode_utf8($opts{process_name});
}

sub _encode_unregister
{
    my (%opts) = @_;
    pack('VVVV',
         3, # API number
         0, # API version
         $opts{pid},
         $opts{handle});
}

sub _encode_submit
{
    my (%opts) = @_;
    pack('VVVVVVVV',
         0, # API number
         2, # API version
         $opts{syncoption},
         $opts{pid},
         $opts{handle},
         bytes::length($opts{location}),
         bytes::length($opts{service}),
         bytes::length($opts{request}))
        . encode_utf8($opts{location})
        . encode_utf8($opts{service})
        . encode_utf8($opts{request});
}

sub _decode_register
{
    my ($in) = @_;
    my ($first_rc, $rc, $handle) = unpack('VVV',$in);
    if (defined($first_rc) && ($first_rc!=0)) {
        return ($first_rc, "Register failed (first_rc=$first_rc)");
    }
    my $got = bytes::length($in);
    my $want = 12;
    if ($got != $want)
    {
        return ($STAF::kCommunicationError, "Truncated response to 'register' (got $got, want $want)");
    }
    if ($rc != 0)
    {
        return ($rc, "Register failed (rc=$rc)");
    }
    return (0,'',$handle);
}

sub _decode_unregister
{
    my ($in) = @_;

    my ($first_rc, $rc) = unpack('VV',$in);
    if (defined($first_rc) && ($first_rc!=0))
    {
        return ($first_rc, "UnRegister failed (first_rc=$first_rc)");
    }
    my $got = bytes::length($in);
    my $want = 8;
    if ($got != $want)
    {
        return ($STAF::kCommunicationError, "Truncated response to 'unregister' (got $got, want $want)");
    }
    if ($rc != 0)
    {
        return ($rc, 'UnRegister failed (rc='.$rc.')');
    }
    return (0, '');
}

sub _decode_submit
{
    my ($in) = @_;

    my ($first_rc, $rc, $result_len, $result_bytes) = unpack('VVVa*',$in);
    if (defined($first_rc) && ($first_rc!=0))
    {
        return ($first_rc, 'Submit failed (first_rc='.$first_rc.')');
    }
    my $got = bytes::length($in);
    my $want = 12 + ($result_len || 0);
    if ($got != $want)
    {
        return ($STAF::kCommunicationError, "Truncated response to 'submit' (got $got, want $want)");
    }
    my $result = decode_utf8($result_bytes);
    return ($rc, $result);
}

my $_staf_temp_dir      = '/tmp';
my $_staf_instance_name = 'STAF';

sub _transact
{
    my (%opts) = @_;

    if (!defined($_staf_temp_dir))
    {
        $_staf_temp_dir = $ENV{STAF_TEMP_DIR} || '/tmp';
    }

    if (!defined($_staf_instance_name))
    {
        $_staf_instance_name = $ENV{STAF_INSTANCE_NAME} || 'STAF';
    }

    my $path = $_staf_temp_dir . '/STAFIPC_' . $_staf_instance_name;
    my $fh = IO::Socket::UNIX->new($path) ||
             return ($STAF::kCommunicationError,
                     "Could not connect to STAF local IPC socket '$path': $^E");

    my $out = $opts{encoder}->(%opts);

    my $pos = 0;
    while ($pos < length($out)) {
        my $retval = $fh->syswrite($out,length($out),$pos);
        if (!defined($retval) || $retval<0)
        {
            return ($STAF::kCommunicationError, "Error while sending request: $^E");
        }
        elsif ($retval>0)
        {
            $pos+=$retval;
        }
        else
        {
            return ($STAF::kCommunicationError, "Unexpected EOF while sending request: $^E");
        }
    }

    my $in = '';
    while (1)
    {
        my $retval = $fh->sysread($in,BUFSIZE,length($in));
        if (!defined($retval) || $retval < 0)
        {
            return ($STAF::kCommunicationError, "Error while reading reply: $^E");
        }
        last if ($retval==0);
    }

    return $opts{decoder}->($in);
}

1;

__END__
